"""
Tests for class FileXXX
"""
import unittest

from generalfile.base.classfile import File
from test.base.setUpWorkDir import SetUpWorkDir

class FileXXXTest(unittest.TestCase):
    def setUp(self):
        """Set working dir and clear folder"""
        SetUpWorkDir.activate()

    def test__write_XXX(self):
        pass

    def test__read_XXX(self):
        pass

